/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 2002-2006
 *      Sleepycat Software.  All rights reserved.
 *
 * $Id: FileReader.java,v 1.94 2006/01/03 21:55:49 bostic Exp $
 */

package com.sleepycat.je.log;

import java.io.EOFException;
import java.io.IOException;
import java.nio.Buffer;
import java.nio.ByteBuffer;

import com.sleepycat.je.mock.DatabaseException;
import com.sleepycat.je.mock.DbLsn;
import com.sleepycat.je.mock.EnvironmentImpl;
import com.sleepycat.je.mock.LogUtils;
import com.sleepycat.je.checksum.ChecksumValidator;

/**
 * A FileReader is an abstract class that traverses the log files, reading in
 * chunks of the file at a time. Concrete subclasses perform a particular action
 * to each entry.
 */
public class FileReader exhibits 
		FileReaderInitializationJP,
		FileReaderValidateChecksumJP,
		FileReaderDeterminIsCollectDataJP,
		FileReaderReadHeaderJP,
		FileReaderStartChecksumJP{
	
	pointcut FileReaderInitializationJP : 
		execution(new(..)) && this(_this) && args(env,..);

	public EnvironmentImpl env;
	protected FileManager fileManager;

	/* Buffering reads */
	private ByteBuffer readBuffer;
	private ByteBuffer saveBuffer;
	private int maxReadBufferSize;

	/* Managing the buffer reads */
	private boolean singleFile;
	protected boolean eof;

	private boolean forward;

	/*
	 * ReadBufferFileNum, readBufferFileStart and readBufferFileEnd indicate how
	 * the read buffer maps to the file. For example, if the read buffer size is
	 * 100 and the read buffer was filled from file 9, starting at byte 100,
	 * then readBufferFileNum = 9 readBufferFileStart = 100 readBufferFileEnd =
	 * 200
	 */
	public long readBufferFileNum;
	protected long readBufferFileStart;
	protected long readBufferFileEnd;

	/* stats */
	private int nRead;

	/*
	 * The number of times we've tried to read in a log entry that was too large
	 * for the read buffer.
	 */
	private long nRepeatIteratorReads;

	/* Info about the last entry seen. */
	public byte currentEntryTypeNum;
	public byte currentEntryTypeVersion;
	public long currentEntryPrevOffset;
	public int currentEntrySize;
//	protected long currentEntryChecksum;

	/*
	 * nextEntryOffset is used to set the currentEntryOffset after we've read an
	 * entry.
	 */
	public long currentEntryOffset;
	public long nextEntryOffset;
	public long startLsn;
	private long finishLsn;

	/* For checking checksum on the read. */
//	protected ChecksumValidator cksumValidator;
//	boolean doValidateChecksum;
//	private boolean alwaysValidateChecksum;

	/* True if this is the scavenger and we are expecting checksum issues. */
//	protected boolean anticipateChecksumErrors;

	/**
	 * A FileReader just needs to know what size chunks to read in.
	 * 
	 * @param endOfFileLsn
	 *            indicates the end of the log file
	 */
	public FileReader(EnvironmentImpl env, int readBufferSize, boolean forward,
			long startLsn, Long singleFileNumber, long endOfFileLsn,
			long finishLsn) throws IOException, DatabaseException {

		this.env = env;
		this.fileManager = env.getFileManager();
		

		/* Allocate a read buffer. */
		this.singleFile = (singleFileNumber != null);
		this.forward = forward;

		readBuffer = ByteBuffer.allocate(readBufferSize);

		saveBuffer = ByteBuffer.allocate(readBufferSize);

		/* Determine the starting position. */
		this.startLsn = startLsn;
		this.finishLsn = finishLsn;

		/* stats */
		nRead = 0;
	}

	/**
//	 * Whether to always validate the checksum, even for non-target entries.
//	 */
//	public void setAlwaysValidateChecksum(boolean validate) {
//		alwaysValidateChecksum = validate;
//	}

	/**
	 * @return the number of entries processed by this reader.
	 */
	public int getNumRead() {
		return nRead;
	}

	public long getNRepeatIteratorReads() {
		return nRepeatIteratorReads;
	}

	/**
	 * Get LSN of the last entry read.
	 */
	public long getLastLsn() {
		return DbLsn.makeLsn(readBufferFileNum, currentEntryOffset);
	}

	/**
	 * readNextEntry scans the log files until either it's reached the end of
	 * the log or has hit an invalid portion. It then returns false.
	 * 
	 * @return true if an element has been read
	 */
	public boolean readNextEntry() throws DatabaseException, IOException {

		boolean foundEntry = false;
		try {
			while ((!eof) && (!foundEntry)) {

				/* Read the next header. */

				ByteBuffer dataBuffer = readData(LogManager.HEADER_BYTES, true);

				readHeader(dataBuffer);

				boolean isTargetEntry = true;/*
												 * isTargetEntry(currentEntryTypeNum,
												 * currentEntryTypeVersion);
												 */
				boolean collectData = hook_isCollectData(isTargetEntry);

				exhibit new FileReaderStartChecksumJP(this,isTargetEntry,dataBuffer){};
				//TODO collectData zurueckschreiben
				
				/*
				 * Read in the body of the next entry. Note that even if this
				 * isn't a targetted entry, we have to move the buffer position
				 * along.
				 */
				dataBuffer = readData(currentEntrySize, collectData);

				/*
				 * We've read an entry. Move up our offsets if we're moving
				 * forward. If we're moving backwards, we set our offset before
				 * we read the header, because we knew where the entry started.
				 */
				if (forward) {
					currentEntryOffset = nextEntryOffset;
					nextEntryOffset += LogManager.HEADER_BYTES
							+ currentEntrySize;
				}

				exhibit new FileReaderValidateChecksumJP(this,isTargetEntry,dataBuffer){};

				if (isTargetEntry) {

					/*
					 * For a target entry, call the subclass reader's
					 * processEntry method to do whatever we need with the
					 * entry. It returns true if this entry is one that should
					 * be returned. Note that some entries, although targetted
					 * and read, are not returned.
					 */
					if (processEntry(dataBuffer)) {
						foundEntry = true;
						nRead++;
					}
				} else if (collectData) {

					/*
					 * For a non-target entry that was validated, the buffer is
					 * positioned at the start of the entry; skip over it.
					 */
					threadSafeBufferPosition(dataBuffer,
							threadSafeBufferPosition(dataBuffer)
									+ currentEntrySize);
				}
			}
		} catch (EOFException e) {
			eof = true;
		} catch (DatabaseException e) {
			eof = true;
			/* Report on error. */

			throw e;
		}
		return foundEntry;
	}
	
	pointcut FileReaderDeterminIsCollectDataJP : execution(boolean hook_isCollectData(..)) && this(_this);
	public boolean hook_isCollectData(boolean isTargetEntry) {
		return isTargetEntry; 
	}

	pointcut FileReaderReadHeaderJP : execution(void readHeader(ByteBuffer)) && args(dataBuffer) && this(_this);
	private void readHeader(ByteBuffer dataBuffer) throws DatabaseException {

		/* Get the checksum for this log entry. */
		dataBuffer.mark();

		/* Read the log entry header. */
		currentEntryTypeNum = dataBuffer.get();

		/*
		 * Always validate the entry type, since this check is cheap. Throw a
		 * DbChecksumException so that LastFileReader and others will recognize
		 * this as data corruption.
		 */

		currentEntryTypeVersion = dataBuffer.get();
		currentEntryPrevOffset = LogUtils.getUnsignedInt(dataBuffer);
		currentEntrySize = dataBuffer.getInt();
	}

//	/**
//	 * Reset the checksum and add the header bytes. This method must be called
//	 * with the entry header data at the buffer mark.
//	 */
//	private void startChecksum(ByteBuffer dataBuffer) throws DatabaseException {
//
//		/* Move back up to the beginning of the cksum covered header. */
//		cksumValidator.reset();
//		int entryStart = threadSafeBufferPosition(dataBuffer);
//		dataBuffer.reset();
//		cksumValidator.update(env, dataBuffer, LogManager.HEADER_CONTENT_BYTES,
//				anticipateChecksumErrors);
//
//		/* Move the data buffer back to where the log entry starts. */
//		threadSafeBufferPosition(dataBuffer, entryStart);
//	}
//
//	/**
//	 * Add the entry bytes to the checksum and check the value. This method must
//	 * be called with the buffer positioned at the start of the entry.
//	 */
//	private void validateChecksum(ByteBuffer entryBuffer)
//			throws DatabaseException {
//
//		cksumValidator.update(env, entryBuffer, currentEntrySize,
//				anticipateChecksumErrors);
//
//	}

	/**
	 * Try to read a specified number of bytes.
	 * 
	 * @param amountToRead
	 *            is the number of bytes we need
	 * @param collectData
	 *            is true if we need to actually look at the data. If false, we
	 *            know we're skipping this entry, and all we need to do is to
	 *            count until we get to the right spot.
	 * @return a byte buffer positioned at the head of the desired portion, or
	 *         null if we reached eof.
	 */
	private boolean first = true;
	boolean foundEntry;

	private ByteBuffer readData(int amountToRead, boolean collectData)
			throws IOException, DatabaseException, EOFException {
		if (first)
			return new MockLogSource(true).getBytes(0);
		else {
			first = false;
			throw new EOFException();
		}

	}

	protected boolean processEntry(ByteBuffer entryBuffer)
			throws DatabaseException {
		return true;
	}

	public int threadSafeBufferPosition(ByteBuffer buffer) {
		while (true) {
			try {
				return buffer.position();
			} catch (IllegalArgumentException IAE) {
				continue;
			}
		}
	}

	public Buffer threadSafeBufferPosition(ByteBuffer buffer, int newPosition) {
		while (true) {
			try {
				return buffer.position(newPosition);
			} catch (IllegalArgumentException IAE) {
				continue;
			}
		}
	}
}
